# -*- coding: utf-8 -*-
# MorDL project: utils
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a bunch of utilities to use with PyTorch.
"""
import sys


BATCH_SIZE = 64
TRAIN_BATCH_SIZE = 32
NONE_TAG = '_'
CONFIG_ATTR = '_mordl_config'
CONFIG_EXT = '.config.json'
LOG_FILE = sys.stdout
CDICT_COEF_THRESH = .99
